import os
import re
from abc import abstractmethod

import pandas as pd
from groq import Groq

from ..base.base import Drag


class Groq_Chat(Drag):
    def __init__(self, config=None):
        Drag.__init__(self, config=config)

        if "api_key" in config:
            self.client = Groq(api_key=config["api_key"])
        if "model" in config:
            self.model = config["model"]
        else:
            self.model = "mixtral-8x7b-32768"

    def submit_prompt(self, prompt, temperature=0.4, **kwargs) -> str:
        if prompt is None:
            raise Exception("Prompt is None")
        if len(prompt) == 0:
            raise Exception("Prompt is empty")
        
        chat_completion = self.client.chat.completions.create(
            messages=prompt,
            model=self.model,
            temperature=temperature,
        )
        llm_response = self.clear_llm_response(llm_response=chat_completion.choices[0].message.content)
        return llm_response
    
    def clear_llm_response(self, llm_response, **kwargs):
        llm_response = self.extract_llm_response_between_triple_backticks(llm_response=llm_response)
        llm_response = self.extract_characters_between_asterisks(llm_response=llm_response)
        llm_response = llm_response.replace("```", "")
        llm_response = llm_response.replace("python", "")
        llm_response = llm_response.strip()
        return llm_response
    
    def extract_llm_response_between_triple_backticks(self, llm_response):
        start_index = llm_response.find('```')
        end_index = llm_response.find('```', start_index + 3)
        if (start_index != -1 or end_index != -1) and start_index != end_index:
            return llm_response[start_index:end_index+3]
        return llm_response
        
    def extract_characters_between_asterisks(self, llm_response):
        start_index = llm_response.find("*")
        end_index = llm_response.rfind("*")
        if (start_index != -1 or end_index != -1) and start_index != end_index:
            return llm_response[start_index+2:end_index-1]
        return llm_response


import sys
import nbformat
from nbformat.v4 import new_code_cell, new_markdown_cell

def check_cell_for_comments(cell: nbformat.notebooknode.NotebookNode, comments: list):
    if cell.cell_type == 'markdown':
        return True
    elif cell.cell_type == 'code':
        return any(comment in line for line in cell.source.split('\n') for comment in comments if line.strip())
    return False

def remove_cells_without_comments(notebook: nbformat.NotebookNode, comments: list):
    new_cells = [cell for cell in notebook.cells if check_cell_for_comments(cell, comments)]
    notebook.cells = new_cells

def main():
    if len(sys.argv) != 2:
        print("Usage: python drag_clean.py <path_to_notebook.ipynb>")
        sys.exit(1)

    ipynb_path = sys.argv[1]
    with open(ipynb_path, 'r') as f:
        notebook = nbformat.read(f, as_version=4)

    comments_to_keep = ['# Code generated by DRAG',
                        '#Code generated by DRAG',
                        '# DRAG',
                        '#DRAG',
                        ]
    remove_cells_without_comments(notebook, comments_to_keep)

    new_ipynb_path = ipynb_path.replace('.ipynb', '_cleaned.ipynb')
    with open(new_ipynb_path, 'w') as f:
        nbformat.write(notebook, f)


if __name__ == "__main__":
    main()


def split_on_plus_minus(string):
    parts = string.split('-')
    if  len(parts) == 1:
        flow = parts[0]
        mods = []
    else:
        flow = parts.pop(0)
        mods = parts 
    return flow, mods


from abc import ABC, abstractmethod

import numpy as np
import pandas as pd
import json
import time
from IPython.core.getipython import get_ipython
from ..utils import *

class Drag(ABC):
    def  __init__(self, config):
        self.config = config

    def data(self, file_path, df_name='df'):
        self.add_cell(f"# Code generated by DRAG\nimport numpy as np\nimport pandas as pd\n\n{df_name} = pd.read_csv('{file_path}')")
        self.add_questions(question=df_name, output=str(pd.read_csv(file_path).head(3)))

    def ask(self, data: dict, question: str, context=True, question_enhancer=False):
        llm_response = self.generate_prompt(data=data, question=question, context=context, question_enhancer=question_enhancer)
        self.add_questions(question=question, output=llm_response)
        self.add_cell(f"# Code generated by DRAG\n{llm_response}")

    def add_cell(self, text, source='set_next_input', replace=False):
        shell = get_ipython()
        payload = dict(
            source=source,
            text=text,
            replace=replace,
        )
        shell.payload_manager.write_payload(payload, single=False)

    def generate_prompt(self, data, question: str, context=True, question_enhancer=False, **kwargs) -> str:
            if type(context) is bool:
                if context:
                    question_sql_list = self.get_similar_question_sql(question, **kwargs)
                    context = question_sql_list[0]
                else:
                    context = None
            if question_enhancer:
                question = self.question_enhancer(question=question, data=data, context=context)
            prompt = self.get_prompt(
                data=data,
                question=question,
                context=context,
                **kwargs,
            )
            #print(f"\n{prompt}\n")
            llm_response = self.submit_prompt(prompt, **kwargs)
            return llm_response

    def get_prompt( self, data: dict, question: str, context=None, **kwargs,):
            data_keys = [k for k in data.keys()]
            data_values = [v for v in data.values()]
            initial_prompt = f"You are an expert data scientist who helps do data analysis and write python code according to the user's request. \
It is very important that you only answer with python code and not with explanations.\n\
The code you generate will be inserted inside a Python notebook, \
in which a Pandas DataFrame called {', '.join(data_keys)} has already been declared."
            data_prompt = f"Consider the following DataFrame {', '.join(data_keys)}:\n"
            data_prompt = self.add_df_to_prompt(initial_prompt=initial_prompt, data=data_values)
            context_prompt = ""
            if context is not None and context != "":
                context_prompt = f"Consider this python code already in the notebook, use it as a basis for generating other code but it is very important not to return it as output. \
It is important to avoid duplicate code.\n'''python\n{context}\n'''"
            prompt = [
                {
                    "role": "system",
                    "content": initial_prompt
                },
                {
                    "role": "system",
                    "content": data_prompt
                },
                {
                    "role": "system",
                    "content": context_prompt
                },
                {
                    "role": "user",
                    "content": f"Only generate python code that allows: {question}"
                }
            ]
            print(f"\n\nSYSTEM:\n{initial_prompt}\n\nUSER:\nGenerates python code that enables:\n{question}")
            return prompt

    def add_df_to_prompt(self, initial_prompt: str, data) -> str:
            for d in data:
                if len(d) > 4:
                    initial_prompt += f"{str(d.head(3))}\n"
                else:
                    initial_prompt += f"{str(d)}\n"
            return initial_prompt
    
    def question_enhancer(self, question, data, context):
        data_keys = [k for k in data.keys()]
        data_values = [v for v in data.values()]
        initial_prompt = f"The user passes you a sentence that will be used as a prompt, you enhance that prompt and return it as output.\
Enhance the sentence considering that it refers to the following DataFrames {', '.join(data_keys)}\
, but it is very important not to include it in the output sentence:\n"
        initial_prompt += self.add_df_to_prompt(initial_prompt=initial_prompt, data=data_values)
        if context is not None:
            initial_prompt += f"\nAlso consider this code, rewrite the prompt so that it can be used as the basis for later generated code:\n{context}\n"
        initial_prompt  += f"\nEnhance the sentence so that the LLM receiving the prompt will be facilitated in carrying out its task.\n\
Return only the improved sentence, it is very important that you do not return anything else."
        prompt = [
                {
                    "role": "system",
                    "content": initial_prompt
                },
                {
                    "role": "user",
                    "content": f"Enhance this prompt:\n{question}"
                }
            ]
        return self.submit_prompt(prompt, temperature=0.2)
    
    def pipe(self, data: dict, question: str, context=False, question_enhancer=True):
        questions = question.split(";")
        cells = []
        llm_response = context
        for q in questions:
            llm_response = self.generate_prompt(data=data, question=q, context=llm_response, question_enhancer=question_enhancer)
            self.add_questions(question=question, output=llm_response)
            cells.append(llm_response)
        for cell in reversed(cells):
            self.add_cell(f"# Code generated by DRAG\n{cell}")

    def flow(self, data: dict, flow: str, context=False, question_enhancer=True):
        flow, mods = split_on_plus_minus(flow)
        with open(f"../src/drag/flows/{flow}.json", 'r') as f:
            flow_steps = json.load(f)
        steps_to_do = {} 
        for idx, key in enumerate(flow_steps['flow']):
            if str(idx) not in mods and key not in mods:
                steps_to_do[key] = flow_steps['flow'][key]
        cells = []
        llm_response = context
        print(steps_to_do)
        for tag, step in steps_to_do.items():
            llm_response = self.generate_prompt(data=data, question=step['Question'], context=llm_response, question_enhancer=question_enhancer)
            self.add_questions(question=step['Question'], output=llm_response)
            cells.append(llm_response)
        for cell in reversed(cells):
            self.add_cell(f"# Code generated by DRAG\n{cell}")

    def flow_steps(self, flow: str):
        with open(f"../src/drag/flows/{flow}.json", 'r') as f:
            flow_steps = json.load(f)
        print(f"{flow}\n{flow_steps['infos']['Description']}")
        for idx, key in enumerate(flow_steps['flow']):
            print(f"\n{idx} - {key}")
            for  k, v in flow_steps['flow'][key].items():
                print(f"    {k}: {v}")


import json
from typing import List
import uuid
from abc import abstractmethod

import chromadb
import pandas as pd
from chromadb.config import Settings
from chromadb.utils import embedding_functions

from ..base.base import Drag

default_ef = embedding_functions.DefaultEmbeddingFunction()


class ChromaDB_VectorStore(Drag):
    def __init__(self, config=None):
        Drag.__init__(self, config=config)

        if config is not None:
            path = config.get("path", ".")
            self.embedding_function = config.get("embedding_function", default_ef)
        else:
            path = "."
            self.embedding_function = default_ef

        self.chroma_client = chromadb.PersistentClient(
            path=path, settings=Settings(anonymized_telemetry=False)
        )
        self.questions_collection = self.chroma_client.get_or_create_collection(
            name="questions", embedding_function=self.embedding_function
        )

    def generate_embedding(self, data: str, **kwargs) -> List[float]:
        embedding = self.embedding_function([data])
        if len(embedding) == 1:
            return embedding[0]
        return embedding

    def add_questions(self, question: str, output: str, **kwargs) -> str:
        question_output_json = json.dumps(
            {
                "question": question,
                "output": output,
            },
            ensure_ascii=False,
        )
        id = str(uuid.uuid4()) + "-question"
        self.questions_collection.add(
            documents=question_output_json,
            embeddings=self.generate_embedding(question_output_json),
            ids=id,
        )
        return id

    def get_training_data(self, **kwargs) -> pd.DataFrame:
        questions_data = self.questions_collection.get()

        df = pd.DataFrame()

        if questions_data is not None:
            # Extract the documents and ids
            documents = [json.loads(doc) for doc in questions_data["documents"]]
            ids = questions_data["ids"]

            # Create a DataFrame
            df_questions = pd.DataFrame(
                {
                    "id": ids,
                    "question": [doc["question"] for doc in documents],
                    "content": [doc["output"] for doc in documents],
                }
            )

            df_questions["training_data_type"] = "question"

            df = pd.concat([df, df_questions])

        return df

    @staticmethod
    def _extract_documents(query_results) -> list:
        """
        Static method to extract the documents from the results of a query.

        Args:
            query_results (pd.DataFrame): The dataframe to use.

        Returns:
            List[str] or None: The extracted documents, or an empty list or single document if an error occurred.
        """
        if query_results is None:
            return []

        if "documents" in query_results:
            documents = query_results["documents"]

            if len(documents) == 1 and isinstance(documents[0], list):
                try:
                    documents = [json.loads(doc) for doc in documents[0]]
                except Exception as e:
                    return documents[0]

            return documents

    def get_similar_questions(self, question: str, **kwargs) -> list:
        return ChromaDB_VectorStore._extract_documents(
            self.questions_collection.query(
                query_texts=[question],
            )
        )